__author__ = 'Kyle'

import math


class Variable:
    """Represent a variable as a class."""

    def __init__(self, coef, exp):
        """Set coefficient and exponent."""
        self.coef = coef
        self.exp = exp

    def __str__(self):
        """Convert variable to a string.

         Returns a string in the form Ax^B.
         """
        if isinstance(self.coef, float):
            if self.coef.is_integer():
                self.coef = int(self.coef)
        if self.coef == 0:
            return "0"
        if self.coef == 1:
            if self.exp == 0:
                return "1"
            elif self.exp == 1:
                return "x"
            return "x^" + str(self.exp)
        if self.exp == 0:
            return str(round(self.coef, 2))
        elif self.exp == 1:
            return str(round(self.coef, 2)) + "x"
        return str(round(self.coef, 2)) + "x^" + str(self.exp)


class Function:
    """Represent a function as a class."""

    def __init__(self, *vars):
        """Set function to the list of variables."""
        self.function = vars

    def __str__(self):
        """Convert the function to a string

        Returns a string in the form Ax^B + Cx^D ...
        """
        function = ""
        for var in self.function:
            function += str(var) + " + "
        return function[:-3]


class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __str__(self):
        return "(" + str(self.x) + ", " + str(self.y) + ")"


def quadratic_formula(a, b, c):
    """Find the zeros of a quadratic function.

    Takes the coefficients of the variables and returns the zeros as a list.
    """
    result = []
    root = math.sqrt(b ** 2 - 4 * a * c)
    top1 = -b + root
    top2 = -b - root
    result.append(top1 / 2 * a)
    result.append(top2 / 2 * a)
    return result


def create_variable(coefficient, exponent):
    """Create a representation of a variable in math.

    Takes a coefficient of a variable and an exponent and creates a representation of a variable with it.
    Returns a Variable object.
    """
    return Variable(coefficient, exponent)


def create_point(x, y):
    """Create a representation of a point.

    Takes an x and y position and creates a representation of a point with it.
    Returns a Point object.
    """
    return Point(x, y)


def create_function(*variables):
    """Create a representation of a function in math.

    Takes several variables and creates a representation of a function with them.
    Returns a Function object.
    """
    return Function(*variables)


def derivative(fn):
    """Find the derivative of a function.

    Takes a Function and returns the derivative of the function.
    """
    dfn = []
    for variable in fn.function:
        if variable.exp == 0:
            continue
        else:
            dfn.append(create_variable(variable.coef * variable.exp, variable.exp - 1))
    return create_function(*dfn)


def antiderivative(dfn):
    """Find the antiderivative of a function.

    Takes a Function and returns the antiderivative of the function without C.
    """
    fn = []
    for variable in dfn.function:
        fn.append(create_variable(variable.coef / (variable.exp + 1), variable.exp + 1))
    return create_function(*fn)


def evaluate(fn, x, precision=2):
    """Find the y value at a point, x.

    Takes a Function and an x value and returns the y value at the point, x.
    """
    total = 0
    for var in fn.function:
        total += var.coef * x ** var.exp
    return round(total, precision)


def integrate(fn, start, stop):
    """Find the area under the curve of a function.

    Takes a Function, a starting point, and a stopping point and returns the area under the curve.
    """
    a_fn = antiderivative(fn)
    return evaluate(a_fn, stop) - evaluate(a_fn, start)


def slope(p1, p2):
    """Find the slope of a line using two points.

    Takes two Points and returns the slope as a float.
    """
    return round((p1.y - p2.y) / (p1.x - p2.x), 2)


def prime_factor(x):
    """Breaks down a number into its prime factors.

    Returns a list of factors for a number, x.
    """
    nums = []
    divisor = 2
    if x >= 0:
        while not x == 1:
            while x % divisor == 0:
                x /= divisor
                nums.append(divisor)
            divisor += 1
    return nums


def fibonacci(n):
    """Calculates the fibonacci numbers up to n.

    Returns a list of fibonacci numbers up to n.
    """
    result = []
    a, b = 0, 1
    while b < n:
        result.append(b)
        a, b = b, a + b
    return result